"""Web client for kakuyomu

This module is a web client for kakuyomu.jp.
"""
import os
import pickle

import requests

from kakuyomu.scrapers.my_page import MyPageScraper
from kakuyomu.scrapers.work_page import WorkPageScraper
from kakuyomu.settings import URL, Config, Login, WorkSettings
from kakuyomu.types import Episode, EpisodeId, LoginStatus, Work, WorkId


class Client:
    """Web client for kakuyomu"""

    session: requests.Session
    cookie: requests.cookies.RequestsCookieJar

    def __init__(self, cookie_path: str = Config.COOKIE):
        """Initialize web client"""
        self.session = requests.Session()
        cookies = self._load_cookie(cookie_path)
        if cookies:
            self.session.cookies = cookies

    def _load_cookie(self, filepath: str) -> requests.cookies.RequestsCookieJar | None:
        cookie: requests.cookies.RequestsCookieJar
        try:
            with open(filepath, "rb") as f:
                cookie = pickle.load(f)
                return cookie
        except FileNotFoundError:
            return None
        except pickle.UnpicklingError:
            return None

    def _get(self, url: str, **kwargs) -> requests.Response:  # type: ignore
        return self.session.get(url, **kwargs)

    def _post(self, url: str, **kwargs) -> requests.Response:  # type: ignore
        return self.session.post(url, **kwargs)

    def status(self) -> LoginStatus:
        """Get login status"""
        res = self._get(URL.MY)
        if res.text.find("ログイン") != -1:
            return LoginStatus(is_login=False, email="")
        else:
            return LoginStatus(is_login=True, email="Config.EMAIL_ADDRESS")

    def logout(self) -> None:
        """Logout"""
        self.session.cookies.clear()
        if os.path.exists(Config.COOKIE):
            os.remove(Config.COOKIE)

    def login(self) -> None:
        """Login"""
        res = self._get(URL.LOGIN)
        email_address = Login.EMAIL_ADDRESS
        password = Login.PASSWORD

        data = {"email_address": email_address, "password": password}
        headers = {"X-requested-With": "XMLHttpRequest"}

        res = self._post(URL.LOGIN, data=data, headers=headers)

        # save cookie to a file
        filepath = Config.COOKIE
        with open(filepath, "wb") as f:
            pickle.dump(res.cookies, f)

    def get_works(self) -> dict[WorkId, Work]:
        """Get works"""
        res = self._get(URL.MY)
        html = res.text
        works = MyPageScraper(html).scrape_works()
        return works

    def get_episodes(self, work_id: WorkId = WorkSettings.ID) -> dict[EpisodeId, Episode]:
        """Get episodes"""
        res = self._get(URL.MY_WORK.format(work_id=work_id))
        html = res.text
        episodes = WorkPageScraper(html).scrape_episodes()
        return episodes



import os

import pytest

from kakuyomu.logger import get_logger
from kakuyomu.client.web import Client

COOKIE_PATH = "/tmp/kakuyomu_cookie"


def set_color() -> None:
    import coloredlogs

    coloredlogs.DEFAULT_LEVEL_STYLES = {
        "critical": {"color": "red", "bold": True},
        "error": {"color": "red"},
        "warning": {"color": "yellow"},
        "notice": {"color": "magenta"},
        "info": {},
        "debug": {"color": "green"},
        "spam": {"color": "green", "faint": True},
        "success": {"color": "green", "bold": True},
        "verbose": {"color": "blue"},
    }
    logger = get_logger()
    coloredlogs.install(level="INFO", logger=logger, fmt="%(asctime)s : %(message)s", datefmt="%Y/%m/%d %H:%M:%S")
    coloredlogs.install(level="DEBUG", logger=logger, fmt="%(asctime)s : %(message)s", datefmt="%Y/%m/%d %H:%M:%S")
    coloredlogs.install(level="WARN", logger=logger, fmt="%(asctime)s : %(message)s", datefmt="%Y/%m/%d %H:%M:%S")


set_color()


def remove_cookie() -> None:
    if os.path.exists(COOKIE_PATH):
        os.remove(COOKIE_PATH)


@pytest.fixture(scope="class")
def client() -> Client:
    remove_cookie()
    client = Client(COOKIE_PATH)
    client.login()
    return client


@pytest.fixture
def login_client() -> Client:
    remove_cookie()
    client = Client(COOKIE_PATH)
    client.login()
    return client


@pytest.fixture
def logout_client() -> Client:
    remove_cookie()
    client = Client(COOKIE_PATH)
    return client



from kakuyomu.client import Client

from .helper import Test


class TestLogin(Test):
    def test_status_not_login(self, logout_client: Client) -> None:
        status = logout_client.status()
        assert not status.is_login

    def test_status_login(self, login_client: Client) -> None:
        login_client.login()
        status = login_client.status()
        assert status.is_login



from kakuyomu.client import Client
from kakuyomu.types import Episode, Work

from .helper import Test

work = Work(
    id="16816927859498193192",
    title="アップロードテスト用",
)

episode = Episode(
    id="16816927859880032697",
    title="第4話",
)


class TestWork(Test):
    def test_work_list(self, client: Client) -> None:
        works = client.get_works()
        assert work.id in works
        assert works[work.id].title == work.title

    def test_episode_list(self, client: Client) -> None:
        episodes = client.get_episodes(work.id)
        assert episode.id in episodes
        assert episodes[episode.id].title == episode.title



"""Kakuyomu CLI

Command line interface for kakuyomu.jp
"""
import click

from kakuyomu.client import Client

client = Client()


@click.group()
def cli() -> None:
    """Kakuyomu CLI

    Command line interface for kakuyomu.jp
    """
    pass


@cli.command()
def status() -> None:
    """Show login status"""
    print(client.status())


@cli.command()
def logout() -> None:
    """Logout"""
    client.logout()
    print("logout")


@cli.command()
def login() -> None:
    """Login"""
    client.login()
    print(client.status())


@cli.command()
def works() -> None:
    """List work titles"""
    for work in client.get_works().values():
        print(work)


@cli.command()
def episodes() -> None:
    """List episodes titles"""
    for episode in client.get_episodes().values():
        print(episode)


def main() -> None:
    """CLI entry point"""
    cli()


if __name__ == "__main__":
    main()



"""Scrape work page."""

import bs4

from kakuyomu.types import Episode, EpisodeId


class WorkPageScraper:
    """Class for scrape work page."""

    html: str

    def __init__(self, html: str):
        """Initialize WorkPageScraper"""
        self.html = html

    def scrape_episodes(self) -> dict[EpisodeId, Episode]:
        """Scrape episodes from work page"""
        soup = bs4.BeautifulSoup(self.html, "html.parser")
        links = soup.select("td.episode-title a")
        result: dict[EpisodeId, Episode] = {}
        for link in links:
            href = link.get("href")
            if not href or not isinstance(href, str):
                continue
            episode_id = href.split("/")[-1]
            episode_title = link.text
            episode = Episode(id=episode_id, title=episode_title)
            result[episode_id] = episode
        return result



"""CLI Settings"""
import os
from functools import lru_cache
from typing import Any, Final

import toml

from kakuyomu.logger import get_logger
from kakuyomu.types import WorkConfig

logger = get_logger()
config_dirname = ".kakuyomu"
work_config_file_name = "config.toml"


def find_work_dir() -> str:
    """Find work dir

    Find work dir from current working directory.
    """
    cwd = os.getcwd()
    while True:
        path = os.path.join(cwd, config_dirname)
        if os.path.exists(path):
            if os.path.isdir(path):
                logger.info(f"work dir found: {cwd}")
                return cwd
        cwd = os.path.dirname(cwd)
        if cwd == "/":
            raise FileNotFoundError(f"{config_dirname} not found")


def find_config_dir() -> str:
    """Find config_dir

    Find config_dir from work dir.
    """
    root = find_work_dir()
    return os.path.join(root, config_dirname)


class ConstMeta(type):
    """Const meta class"""

    def __setattr__(self, name: str, value: Any) -> None:
        """Deny reassign const"""
        if name in self.__dict__:
            raise TypeError(f"cannot reassign const {name!r}")
        else:
            self.__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        """Deny delete const"""
        if name in self.__dict__:
            raise TypeError(f"cannot delete const {name!r}")


class URL(metaclass=ConstMeta):
    """URL constants"""

    ROOT: Final[str] = "https://kakuyomu.jp"
    LOGIN: Final[str] = f"{ROOT}/login"
    MY: Final[str] = f"{ROOT}/my"
    MY_WORK: Final[str] = f"{MY}/works/" + "{work_id}"
    ANNTENA_WORKS: Final[str] = f"{ROOT}/anntena/works"


class Login(metaclass=ConstMeta):
    """Login constants"""

    EMAIL_ADDRESS: Final[str] = os.environ.get("KAKUYOMU_EMAIL_ADDRESS", "")
    PASSWORD: Final[str] = os.environ.get("KAKUYOMU_PASSWORD", "")


class Config(metaclass=ConstMeta):
    """Config constants"""

    DIR: Final[str] = find_config_dir()
    COOKIE: Final[str] = os.path.join(DIR, "cookie")


@lru_cache(maxsize=5)
def work_config(config_dir: str = Config.DIR) -> WorkConfig:
    """Load work config

    Load work config from config_dir.
    Result is caches.
    """
    work_file = os.path.join(config_dir, work_config_file_name)
    with open(work_file, "r") as f:
        return WorkConfig(**toml.load(f))


class WorkSettings(metaclass=ConstMeta):
    """Work settings"""

    ID: Final[str] = work_config().id


os.makedirs(Config.DIR, exist_ok=True)

if not Login.PASSWORD or not Login.EMAIL_ADDRESS:
    raise ValueError("KAKUYOMU_EMAIL_ADDRESS and KAKUYOMU_PASSWORD must be set")



"""Scrape my page."""
import bs4

from kakuyomu.types import Work, WorkId


class MyPageScraper:
    """Class for scrape my page."""

    html: str

    def __init__(self, html: str):
        """Initialize MyPageScraper"""
        self.html = html

    def scrape_works(self) -> dict[WorkId, Work]:
        """Scrape works from my page"""
        soup = bs4.BeautifulSoup(self.html, "html.parser")
        links = soup.find_all("h2", class_="workColumn-workTitle")
        result = {}
        for link in links:
            work_id = link.a.get("href").split("/")[-1]
            work_title = link.a.text
            work = Work(id=work_id, title=work_title)
            result[work_id] = work
        return result



"""Define type aliases and models."""

from typing import TypeAlias

from pydantic import BaseModel

WorkId: TypeAlias = str
EpisodeId: TypeAlias = str


class Work(BaseModel):
    """Work model"""

    id: WorkId
    title: str


class Episode(BaseModel):
    """Episode model"""

    id: WorkId
    title: str


class LoginStatus(BaseModel):
    """Login status model"""

    is_login: bool
    email: str


class WorkConfig(BaseModel):
    """Work config model"""

    id: WorkId



"""Module for logging"""
import logging

import coloredlogs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
coloredlogs.install(logger=logger)


def get_logger() -> logging.Logger:
    """Get logger"""
    return logger
